/*
 * Copyright (c) 1999, 2000, 2001, 2002, 2003, 2005, 2010 Greg Haerr <greg@censoft.com>
 * Portions Copyright (c) 2002 by Koninklijke Philips Electronics N.V.
 *
 * Win32 API upper level graphics drawing routines
 */
#include "windows.h"
#include "wintern.h"
#include "device.h"
#include <stdlib.h>
#include <string.h>
#include <Protocol/GraphicsOutput.h>
#include <Protocol/HiiFont.h>
#include <Library/UefiHiiServicesLib.h>

typedef struct _FONT_WIDTH_NODE {
  UINT16     FontSize;
  UINT8      FontWidth[0x100];
} FONT_WIDTH_NODE;

typedef struct _FONT_WIDTH_LIST {
  UINT32           NodeCount;
  FONT_WIDTH_NODE  *FontWidthNode;
} FONT_WIDTH_LIST;

UINT16                        mFontSize;
EFI_FONT_DISPLAY_INFO         *mStringInfo = NULL;


EFI_STATUS
TrueTypeStringToImage (
  IN CHAR16                          *String,
  IN EFI_GRAPHICS_OUTPUT_BLT_PIXEL   *Blt,
  IN UINT16                          FontSize,
  IN UINT16                          Width,
  IN UINT16                          Height,
  IN UINT16                          Delta,
  IN EFI_GRAPHICS_OUTPUT_BLT_PIXEL   FgColor,
  IN EFI_GRAPHICS_OUTPUT_BLT_PIXEL   BkColor,
  OUT UINT16                         *StringWidth
  );

#define OLD_DRAWTEXT	0	/* use old mwDrawText routine, deprecated*/

#define MAXSYSCOLORS	29	/* # of COLOR_* system colors*/
#define MAXSTOCKOBJECTS	20	/* # of stock objects*/

#if ERASEMOVE
BOOL mwERASEMOVE = TRUE;	/* default XORMOVE repaint algorithm*/
#else
BOOL mwERASEMOVE = FALSE;	/* default ERASEMOVE repaint algorithm*/
#endif

/* current encoding for non-wide char text functions*/
long mwTextCoding = MWTF_UTF8;	/* usually MWTF_ASCII or MWTF_UTF8*/

/* cast a pointer to an integer*/
#if DOS_TURBOC | MACOSX
#define PTRTOINT	unsigned long
#else
#define PTRTOINT	unsigned int
#endif

static HDC	cliphdc;	/* current window cliprects*/

/* default bitmap for new DCs*/
static MWBITMAPOBJ default_bitmap = {
	{OBJ_BITMAP, TRUE}, 1, 1, 1, 1, 1, 1
};

#if 0
static BOOL MwExtTextOut(HDC hdc, int x, int y, UINT fuOptions,
		CONST RECT *lprc, LPCVOID lpszString, UINT cbCount,
		CONST INT *lpDx, int flags);
static int MwDrawText(HDC hdc, LPCVOID lpString, int nCount, LPRECT lpRect, UINT uFormat,
	int flags);
static LONG mwTabbedTextOut(HDC hdc, int x, int y, LPCTSTR lpszString, int cbString,
	int ntabs, LPINT lpTabStops, int nTabOrigin, BOOL noDraw, int flags);
#endif

HDC WINAPI
GetDCEx(HWND hwnd,HRGN hrgnClip,DWORD flags)
{
	HDC	hdc;

	if(!hwnd)		/* handle NULL hwnd => desktop*/
		hwnd = rootwp;

	/* handle private DC's*/
	if(hwnd->owndc && !(flags & DCX_WINDOW))
		return hwnd->owndc;

	/* add caching?*/
	hdc = GdItemNew(struct hdc);
	if(!hdc)
		return NULL;

	hdc->psd = &scrdev;
	hdc->hwnd = hwnd;
	if(flags & DCX_DEFAULTCLIP) {
		flags &= ~DCX_DEFAULTCLIP;
		if(hwnd->style & WS_CLIPSIBLINGS)
			flags |= DCX_CLIPSIBLINGS;
		if(hwnd->style & WS_CLIPCHILDREN)
			flags |= DCX_CLIPCHILDREN;
	}
	hdc->flags = flags;
	hdc->bkmode = OPAQUE;
	hdc->textalign = TA_LEFT | TA_TOP | TA_NOUPDATECP;
	hdc->bkcolor = RGB(255, 255, 255);	/* WHITE*/
	hdc->textcolor = RGB(0, 0, 0);		/* BLACK*/
	hdc->brush = (MWBRUSHOBJ *)GetStockObject(WHITE_BRUSH);
	hdc->pen = (MWPENOBJ *)GetStockObject(BLACK_PEN);
	hdc->font = (MWFONTOBJ *)GetStockObject(SYSTEM_FONT);
#if UPDATEREGIONS
	if(hrgnClip) {
		/* make a copy of passed region*/
		hdc->region = (MWRGNOBJ *)CreateRectRgn(0, 0, 0, 0);
		CombineRgn((HRGN)hdc->region, hrgnClip, NULL, RGN_COPY);
	}
#endif

	/* make default bitmap compatible with scrdev
	 * otherwise problems occur later because selecting
	 * in the default bitmap overwrite planes and bpp
	 * in a memory dc, and thus it becomes incompatible
	 * with scrdev.
	 */
	default_bitmap.planes = scrdev.planes;
	default_bitmap.bpp = scrdev.bpp;
	default_bitmap.data_format = scrdev.data_format;
	hdc->bitmap = &default_bitmap;

	hdc->drawmode = R2_COPYPEN;
	hdc->pt.x = 0;
	hdc->pt.y = 0;

	/* assign private DC if CS_OWNDC and not WindowDC*/
	if(hwnd->pClass && (hwnd->pClass->style & CS_OWNDC) && !(flags & DCX_WINDOW)) {
		/* must exclude update region due to BeginPaint GetDCEx call*/
		hdc->flags |= DCX_EXCLUDEUPDATE;
		hwnd->owndc = hdc;
	}

	return hdc;
}

HDC WINAPI
GetDC(HWND hwnd)
{
	/*
	 * Exclude update regions when drawing with GetDC.
	 * This is required because some programs use GetDC
	 * when painting outside of BeginPaint/EndPaint, and
	 * the update region is empty then.
	 */
	return GetDCEx(hwnd, NULL, DCX_DEFAULTCLIP|DCX_EXCLUDEUPDATE);
}

HDC WINAPI
GetWindowDC(HWND hwnd)
{
	/*
	 * Exclude update region for now, since we
	 * don't keep track of non-client update regions yet
	 */
	return GetDCEx(hwnd, NULL,DCX_WINDOW|DCX_DEFAULTCLIP|DCX_EXCLUDEUPDATE);
}

/* free a DC allocated from GetDC*/
int WINAPI
ReleaseDC(HWND hwnd, HDC hdc)
{
	/* don't delete a memory dc on release*/
	if(!hdc || (hdc->psd->flags&PSF_MEMORY))
		return 0;

	if(hdc == cliphdc)
		cliphdc = NULL;

	/* handle private DC's*/
	if(hdc->hwnd->owndc && !(hdc->flags & DCX_WINDOW))
		return 1;

	DeleteObject((HBRUSH)hdc->brush);
	DeleteObject((HPEN)hdc->pen);
#if 0 /* don't delete font resources on ReleaseDC... use DeleteObject instead*/
	DeleteObject((HFONT)hdc->font);
#endif
	DeleteObject((HRGN)hdc->region);
	/*
	 * We can only select a bitmap in a memory DC,
	 * so bitmaps aren't released except through DeleteDC.
	 */
	DeleteObject((HBITMAP)hdc->bitmap);
	GdItemFree(hdc);
	return 1;
}

/* free a dc allocated from CreateCompatibleDC*/
BOOL WINAPI
DeleteDC(HDC hdc)
{
	/* don't delete a normal dc, only memory dc's*/
	if(!hdc || !(hdc->psd->flags&PSF_MEMORY))
		return 0;

	/* free allocated memory screen device*/
	hdc->psd->FreeMemGC(hdc->psd);

	/* make it look like a GetDC dc, and free it*/
	hdc->psd = &scrdev;
	return ReleaseDC(NULL, hdc);
}

void
MwPaintNCArea(HWND hwnd)
{
	SendMessage(hwnd, WM_NCPAINT, 0, 0L);

	/* for now, we always paint NC scrollbar areas*/
	//MwPaintNCScrollbars(hwnd, NULL);
}

HDC WINAPI
BeginPaint(HWND hwnd, LPPAINTSTRUCT lpPaint)
{
	HDC	hdc;

	/* first update non-client area*/
	if(mwforceNCpaint || hwnd->paintNC != mwpaintNC) {
		MwPaintNCArea(hwnd);
		hwnd->paintNC = mwpaintNC;
	}

	/* If ERASEMOVE:
	 * Don't allow windows to repaint while user is moving
	 * a window.  Instead, just erase backgrounds
	 * and indicate delayed painting required, which
	 * will occur after user completes window move.
	 */
	if(mwERASEMOVE && dragwp) {
		hdc = NULL;
		lpPaint->fErase = !DefWindowProc(hwnd, WM_ERASEBKGND, 0, 0L);
		hwnd->gotPaintMsg = PAINT_DELAYPAINT;
	} else {
		//HideCaret(hwnd);

		/* FIXME: mdemo requires update excluded or draw errors occur*/
		//hdc = GetDCEx(hwnd, NULL, DCX_DEFAULTCLIP
		//		|DCX_EXCLUDEUPDATE);	/* FIXME - bug*/
		hdc = GetDCEx(hwnd, NULL, DCX_DEFAULTCLIP);
		/* erase client background, always w/alpha blending*/
		//if(hwnd->nEraseBkGnd > 0 || mwforceNCpaint)
		//	lpPaint->fErase = !SendMessage(hwnd, WM_ERASEBKGND, (WPARAM)hdc, 0L);
		//else
		//	lpPaint->fErase = 0;

		//hwnd->nEraseBkGnd = 0;
	}
	lpPaint->hdc = hdc;

	if( hwnd->paintBrush != NULL )
		DeleteObject ( hwnd->paintBrush );
	if( hwnd->paintPen != NULL )
		DeleteObject ( hwnd->paintPen );
	hwnd->paintBrush = NULL;
	hwnd->paintPen = NULL;

	GetUpdateRect(hwnd, &lpPaint->rcPaint, FALSE);
	return hdc;
}

BOOL WINAPI
EndPaint(HWND hwnd, CONST PAINTSTRUCT *lpPaint)
{
  if( hwnd->paintBrush != NULL )
    {
    SelectObject ( lpPaint->hdc, GetStockObject(NULL_BRUSH) );
    DeleteObject ( hwnd->paintBrush );
    hwnd->paintBrush = NULL;
    }
  if( hwnd->paintPen != NULL )
    {
    SelectObject ( lpPaint->hdc, GetStockObject(BLACK_PEN) );
    DeleteObject ( hwnd->paintPen );
    hwnd->paintPen = NULL;
    }
  ReleaseDC(hwnd, lpPaint->hdc);
#if UPDATEREGIONS
	/* don't clear update region until done dragging*/
	if(mwERASEMOVE && !dragwp)
		GdSetRectRegion(hwnd->update, 0, 0, 0, 0);
#endif
	//ShowCaret(hwnd);
	return TRUE;
}

HDWP
BeginDeferWindowPos(int nNumWindows)
{
	return (HDWP)(UINTN)1;
}

HDWP
DeferWindowPos(HDWP hWinPosInfo, HWND hWnd, HWND hWndInsertAfter,
	       int x, int y, int cx, int cy, UINT uFlags)
{
  SetWindowPos(hWnd, hWndInsertAfter, x, y, cx, cy, uFlags);
  return hWinPosInfo;
}

BOOL
EndDeferWindowPos(HDWP hWinPosInfo)
{
	return TRUE;
}

COLORREF WINAPI
SetTextColor(HDC hdc, COLORREF crColor)
{
	COLORREF	oldtextcolor;

	if (!hdc)
		return CLR_INVALID;
	oldtextcolor = hdc->textcolor;
	hdc->textcolor = (MWCOLORVAL)crColor;
	return oldtextcolor;
}

COLORREF WINAPI
GetTextColor(HDC hdc)
{
	if (!hdc)
		return CLR_INVALID;
	return hdc->textcolor;
}

COLORREF WINAPI
SetBkColor(HDC hdc, COLORREF crColor)
{
	COLORREF	oldbkcolor;

	if (!hdc)
		return CLR_INVALID;
	oldbkcolor = hdc->bkcolor;
	hdc->bkcolor = crColor;
	return oldbkcolor;
}

int WINAPI
SetBkMode(HDC hdc, int iBkMode)
{
	int	oldbkmode;

	if(!hdc)
		return 0;
	oldbkmode = hdc->bkmode;
	hdc->bkmode = iBkMode;
	return oldbkmode;
}

UINT WINAPI
SetTextAlign(HDC hdc, UINT fMode)
{
	UINT	oldfMode;

	oldfMode = hdc->textalign;
	hdc->textalign = fMode;
	return oldfMode;
}

UINT WINAPI
GetTextAlign(HDC hdc)
{
	if(!hdc)
		return GDI_ERROR;
	return hdc->textalign;
}

/* FIXME: releasing a DC does NOT change back the drawing mode!*/
int WINAPI
SetROP2(HDC hdc, int fnDrawMode)
{
	int	newmode, oldmode;

	if(!hdc || (fnDrawMode <= 0 || fnDrawMode > R2_LAST))
		return 0;

	oldmode = hdc->drawmode;
	newmode = fnDrawMode - 1;	/* map to MWROP_xxx*/
	hdc->drawmode = newmode;
	GdSetMode(newmode);
	return oldmode;
}

/*
 * Setup clip region from device context's associated window or bitmap.
 * Memory DC's are always associated with the desktop window, and are
 * always visible.  Return the DC's hwnd if window is visible.
 */
HWND
MwPrepareDC(HDC hdc)
{
	HWND	hwnd;

	if(!hdc || !hdc->hwnd)
		return NULL;

	hwnd = hdc->hwnd;
	if (hwnd->unmapcount && !(hdc->psd->flags & PSF_MEMORY))
		return NULL;

	/*
	 * If the window is not the currently clipped one, then
	 * make it the current one and define its clip rectangles.
	 */
	if(hdc != cliphdc) {
		/* clip memory dc's to the bitmap size*/
		if(hdc->psd->flags&PSF_MEMORY) {
#if DYNAMICREGIONS
			/* If hdc has a clip region, use it! */
			if (hdc->region != NULL && hdc->region->rgn != NULL
			    && hdc->region->rgn->size != 0) {
				LPRECT rptr = &(hdc->region->rgn->extents);
				MWCOORD left, top;
				MWCLIPREGION *tmp;
				/* Bound left,top to 0,0 if they are negative
				 * to avoid an assertion later.
				 */
				left = (rptr->left <= 0) ? 0 : rptr->left;
				top = (rptr->top <= 0) ? 0 : rptr->top;
				//GdSetClipRegion(hdc->psd,
				//	GdAllocRectRegion(rptr->left, rptr->top,
				//		rptr->right, rptr->bottom));

				tmp = GdAllocRegion ();
				GdCopyRegion (tmp, hdc->region->rgn);
				GdSetClipRegion (hdc->psd, tmp);
			} else {
				GdSetClipRegion(hdc->psd,
					GdAllocRectRegion(0, 0,
						hdc->psd->xvirtres, hdc->psd->yvirtres));
			}
#else
			static MWCLIPRECT crc = {0, 0, 0, 0};

			/* If hdc has a clip region, use it! */
			if (hdc->region != NULL && hdc->region->rgn != NULL
			    && hdc->region->rgn->size != 0) {
				LPRECT rptr = &(hdc->region->rgn->extents);

				/* Bound left,top to 0,0 if they are negative
				 * to avoid an assertion later.
				 */
				if (rptr->left > 0)
					crc.x = rptr->left;
				if (rptr->top > 0)
					crc.y = rptr->top;
				crc.width = rptr->right;
				crc.height = rptr->bottom;
			} else {
				crc.width = hdc->psd->xvirtres;
				crc.height = hdc->psd->yvirtres;
			}
			GdSetClipRects(hdc->psd, 1, &crc);
#endif
		} else MwSetClipWindow(hdc);
		cliphdc = hdc;
	}

	return hwnd;
}

/* return RGB value at specified coordinates*/
COLORREF WINAPI
GetPixel(HDC hdc, int x, int y)
{
	HWND		hwnd;
	POINT		pt;
	MWPIXELVALHW pixel;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return CLR_INVALID;
	pt.x = x;
	pt.y = y;
	if(MwIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	/* read pixel value*/
	GdReadArea(hdc->psd, pt.x, pt.y, 1, 1, &pixel);

	return GdGetColorRGB(hdc->psd, pixel);
}

COLORREF WINAPI
SetPixel(HDC hdc, int x, int y, COLORREF crColor)
{
	HWND		hwnd;
	POINT		pt;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return 0;	/* doesn't return previous color*/
	pt.x = x;
	pt.y = y;
	if(MwIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	/* draw point in passed color*/
	GdSetForegroundColor(hdc->psd, crColor);
	GdPoint(hdc->psd, pt.x, pt.y);
	return 0;		/* doesn't return previous color*/
}

BOOL WINAPI
MoveToEx(HDC hdc, int x, int y, LPPOINT lpPoint)
{
	if(!hdc)
		return FALSE;
	if(lpPoint)
		*lpPoint = hdc->pt;
	hdc->pt.x = x;
	hdc->pt.y = y;
	return TRUE;
}

BOOL WINAPI
LineTo(HDC hdc, int x, int y)
{
	HWND		hwnd;
	POINT		beg, end;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	beg.x = hdc->pt.x;
	beg.y = hdc->pt.y;
	end.x = x;
	end.y = y;
	if(MwIsClientDC(hdc)) {
		ClientToScreen(hwnd, &beg);
		ClientToScreen(hwnd, &end);
	}

	/* draw line in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForegroundColor(hdc->psd, hdc->pen->color);
		/* don't draw last point*/
		GdLine(hdc->psd, beg.x, beg.y, end.x, end.y, FALSE);
	}
	hdc->pt.x = x;
	hdc->pt.y = y;
	return TRUE;
}

/* draw line segments by connecting passed points*/
BOOL WINAPI
Polyline(HDC hdc, CONST POINT *lppt, int cPoints)
{
	HWND		hwnd;
	POINT		beg, end;

	if(cPoints <= 1)
		return FALSE;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	if(hdc->pen->style == PS_NULL)
		return TRUE;

	/* draw line in current pen color*/
	GdSetForegroundColor(hdc->psd, hdc->pen->color);

	beg = *lppt++;
	if(MwIsClientDC(hdc))
		ClientToScreen(hwnd, &beg);
	while(--cPoints > 0) {
		end = *lppt++;
		if(MwIsClientDC(hdc))
			ClientToScreen(hwnd, &end);

		/* don't draw last point*/
		GdLine(hdc->psd, beg.x, beg.y, end.x, end.y, FALSE);

		beg = end;
	}
	return TRUE;
}

BOOL WINAPI
Rectangle(HDC hdc, int nLeft, int nTop, int nRight, int nBottom)
{
	HWND	hwnd;
	RECT	rc;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	SetRect(&rc, nLeft, nTop, nRight, nBottom);
	if(MwIsClientDC(hdc))
		MapWindowPoints(hwnd, NULL, (LPPOINT)&rc, 2);

	/* draw rectangle in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForegroundColor(hdc->psd, hdc->pen->color);
		GdRect(hdc->psd, rc.left, rc.top,
			rc.right - rc.left, rc.bottom - rc.top);
	}

	/* fill rectangle in current brush color*/
	if(hdc->brush->style != BS_NULL) {
		InflateRect(&rc, -1, -1);
		GdSetForegroundColor(hdc->psd, hdc->brush->color);
		GdFillRect(hdc->psd, rc.left, rc.top, rc.right - rc.left,
			rc.bottom - rc.top);
	}

	return TRUE;
}
#if 0
BOOL WINAPI
Ellipse(HDC hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect)
{
	HWND	hwnd;
	int	rx, ry;
	RECT	rc;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	SetRect(&rc, nLeftRect, nTopRect, nRightRect, nBottomRect);
	if(MwIsClientDC(hdc))
		MapWindowPoints(hwnd, NULL, (LPPOINT)&rc, 2);

	rx = (rc.right - rc.left)/2 - 1;
	ry = (rc.bottom - rc.top)/2 - 1;
	rc.left += rx;
	rc.top += ry;

	/* fill ellipse in current brush color*/
	if(hdc->brush->style != BS_NULL) {
		InflateRect(&rc, -1, -1);
		GdSetForegroundColor(hdc->psd, hdc->brush->color);
		GdEllipse(hdc->psd, rc.left, rc.top, rx, ry, TRUE);
	}

	/* draw ellipse outline in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForegroundColor(hdc->psd, hdc->pen->color);
		GdEllipse(hdc->psd, rc.left, rc.top, rx, ry, FALSE);
	}

	return TRUE;
}

static void
dopiearc(HDC hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect,
	int ax, int ay, int bx, int by, int type)
{
	HWND	hwnd;
	int	rx, ry;
	RECT	rc, rc2;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return;

	SetRect(&rc, nLeftRect, nTopRect, nRightRect, nBottomRect);
	SetRect(&rc2, ax, ay, bx, by);
	if(MwIsClientDC(hdc)) {
		MapWindowPoints(hwnd, NULL, (LPPOINT)&rc, 2);
		MapWindowPoints(hwnd, NULL, (LPPOINT)&rc2, 2);
	}

	rx = (rc.right - rc.left)/2 - 1;
	ry = (rc.bottom - rc.top)/2 - 1;
	rc.left += rx;
	rc.top += ry;

	/* fill ellipse in current brush color*/
	if(hdc->brush->style != BS_NULL && type == MWPIE) {
		GdSetForegroundColor(hdc->psd, hdc->brush->color);
		GdArc(hdc->psd, rc.left, rc.top, rx, ry,
			rc2.left, rc2.top, rc2.right, rc2.bottom, MWPIE);
	}

	/* draw ellipse outline in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForegroundColor(hdc->psd, hdc->pen->color);
		if(type == MWPIE)
			type = MWARC;	/* MWARCOUTLINE?*/
		GdArc(hdc->psd, rc.left, rc.top, rx, ry,
			rc2.left, rc2.top, rc2.right, rc2.bottom, type);
	}
}

BOOL WINAPI
Arc(HDC hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect,
	int nXStartArc, int nYStartArc, int nXEndArc, int nYEndArc)
{
	dopiearc(hdc, nLeftRect, nTopRect, nRightRect, nBottomRect,
		nXStartArc, nYStartArc, nXEndArc, nYEndArc, MWARC);
	return TRUE;
}

BOOL WINAPI
Pie(HDC hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect,
	int nXRadial1, int nYRadial1, int nXRadial2, int nYRadial2)
{
	dopiearc(hdc, nLeftRect, nTopRect, nRightRect, nBottomRect,
		nXRadial1, nYRadial1, nXRadial2, nYRadial2, MWPIE);
	return TRUE;
}
#endif

BOOL WINAPI
Polygon(HDC hdc, CONST POINT *lpPoints, int nCount)
{
	HWND	hwnd;
	int	i;
	LPPOINT	pp, ppAlloc = NULL;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	if(MwIsClientDC(hdc)) {
		/* convert points to client coords*/
		ppAlloc = (LPPOINT)malloc(nCount * sizeof(POINT));
		if(!ppAlloc)
			return FALSE;
		memcpy(ppAlloc, lpPoints, nCount*sizeof(POINT));
		pp = ppAlloc;
		for(i=0; i<nCount; ++i)
			ClientToScreen(hwnd, pp++);
		pp = ppAlloc;
	} else pp = (LPPOINT)lpPoints;

	/* fill polygon in current brush color*/
	if(hdc->brush->style != BS_NULL) {
		GdSetForegroundColor(hdc->psd, hdc->brush->color);
		GdFillPoly(hdc->psd, nCount, pp);
	}

	/* draw polygon outline in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForegroundColor(hdc->psd, hdc->pen->color);
		GdPoly(hdc->psd, nCount, pp);
	}

	if(ppAlloc)
		free(ppAlloc);
	return TRUE;
}

/* draw nCount polygons*/
BOOL WINAPI
PolyPolygon(HDC hdc, CONST POINT *lpPoints, LPINT lpPolyCounts, int nCount)
{
	while(--nCount >= 0) {
		if (!Polygon(hdc, lpPoints, *lpPolyCounts))
			return FALSE;
		lpPoints += *lpPolyCounts++;
	}
	return TRUE;
}

int WINAPI
FillRect(HDC hdc, CONST RECT *lprc, HBRUSH hbr)
{
	HWND		hwnd;
	RECT 		rc;
	MWBRUSHOBJ *	obr = (MWBRUSHOBJ *)hbr;
	COLORREF	crFill;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd || !obr)
		return FALSE;

	if(!lprc) {
		if(MwIsClientDC(hdc))
			GetClientRect(hwnd, &rc);
		else
			GetWindowRect(hwnd, &rc);
		lprc = &rc;
	} else
		rc = *lprc;
	if(MwIsClientDC(hdc))
		MapWindowPoints(hwnd, NULL, (LPPOINT)&rc, 2);

	/* handle COLOR_xxx + 1 passed as HBRUSH*/
	if((UINTN)obr <= MAXSYSCOLORS)
		crFill = GetSysColor((UINT32)(UINTN)obr-1);
	else {
		/* get color from passed HBRUSH*/
		if(obr->style == BS_NULL)
			return TRUE;
		crFill = obr->color;
	}

	/* fill rectangle in passed brush color*/
	GdSetForegroundColor(hdc->psd, crFill);
	GdFillRect(hdc->psd, rc.left, rc.top,
		rc.right - rc.left, rc.bottom - rc.top);
	return TRUE;
}

int
WINAPI
MwFastFillRect (
  IN HDC        hDC,
  IN CONST RECT *lprc,
  IN COLORREF   crColor
  )
{
  COLORREF  crOld;
  RECT      rc;
  HWND      hwnd;

  if (hDC == NULL) {
    return FALSE;
  }

  crOld = SetBkColor(hDC, crColor);
  rc    = *lprc;
  hwnd  = MwPrepareDC(hDC);

  if(MwIsClientDC(hDC)) {
    MapWindowPoints(hwnd, NULL, (LPPOINT)&rc, 2);
  }

  /* fill rectangle with current background color*/
  GdSetForegroundColor(hDC->psd, hDC->bkcolor);
  GdFillRect (hDC->psd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top);

  SetBkColor(hDC, crOld);

  return TRUE;
}



/* set current input coding*/
void WINAPI
MwSetTextCoding(long mode)
{
	mwTextCoding = mode;
}


/* ascii or utf8 */
BOOL WINAPI
TextOut (HDC Hdc, int x, int y, LPCWSTR String, int cbString)
{
  HWND      hwnd;
  POINT     pt;
  STATIC HDC       hdcMem = NULL;
  HBITMAP         hbmp;
  CHAR16          savechar;
  INT32           Index;
  SIZE            Size;
  STATIC SIZE     MaxSize = {0};
  INT32     StringCount;
  UINTN  FontSize;
  EFI_STATUS      Status;
  UINT16          StringWidth;
//EFI_IMAGE_OUTPUT      *Blt;
//EFI_FONT_DISPLAY_INFO FontInfo;


  hbmp      = NULL;
  savechar  = '\0';


  pt.x = x;
  pt.y = y;

  hwnd = MwPrepareDC(Hdc);
  if(!hwnd) {
    return FALSE;
  }

  if (String[0] == '\0') {
    return TRUE;
  }

  if(MwIsClientDC(Hdc)) {
    ClientToScreen (hwnd, &pt);
  }

  //
  // process string count and narrow_char / wide_char
  //
  StringCount = cbString;
  if (StringCount != -1) {
    savechar = String[cbString];
    *(LPWSTR)(String + cbString) = '\0';
  } else {
    StringCount = (INT32)wcslen ((CHAR16 *)String);
    for (Index = 0; Index < StringCount; Index++) {
      if (String[Index] == 0x0d) {
        *(CHAR16 *)&String[Index] = ' ';
      }
    }
  }

  FontSize = Hdc->font->FontSize;
  if (FontSize == 0) {
    FontSize = 19;
  }
  Size.cx = (INT32)(FontSize * StringCount);
  Size.cy = (INT32) FontSize;
  if (Size.cx == 0) {
    *(LPWSTR)(String + cbString) = savechar;
    return FALSE;
  }

  if (MaxSize.cx < Size.cx || MaxSize.cy < Size.cy || hdcMem == NULL) {
    MaxSize.cx = MWMAX (MaxSize.cx, Size.cx);
    MaxSize.cy = MWMAX (MaxSize.cy, Size.cy);
    DeleteDC (hdcMem);
    hdcMem   = CreateCompatibleDC(NULL);
    hbmp     = CreateCompatibleBitmap(hdcMem, MaxSize.cx, MaxSize.cy);
    SelectObject(hdcMem, hbmp);
  }
  TrueTypeStringToImage (
    (CHAR16 *) String,
    (EFI_GRAPHICS_OUTPUT_BLT_PIXEL *) hdcMem->psd->addr,
    (UINT16) FontSize,
    (UINT16) Size.cx,
    (UINT16) Size.cy,
    (UINT16) MaxSize.cx,
    *(EFI_GRAPHICS_OUTPUT_BLT_PIXEL *)&Hdc->textcolor,
    *(EFI_GRAPHICS_OUTPUT_BLT_PIXEL *)&Hdc->bkcolor,
    &StringWidth
    );


  Status = EFI_SUCCESS;

  if (!EFI_ERROR (Status)) {
    pt.x = x;
    pt.y = y;
    BitBlt(Hdc, pt.x, pt.y, (INT32) StringWidth, Size.cy, hdcMem, 0, 0, MWROP_SRC_OVER);
  }

  if (cbString != -1) {
    *(LPWSTR)(String + cbString) = savechar;
  }


  return TRUE;

}

#if 0
/* UC16 */
BOOL WINAPI
TextOutW(HDC hdc, int x, int y, LPCWSTR lpszString, int cbString)
{
	/* kaffe port wants MWTF_UTF8 here...*/
	return MwExtTextOut(hdc, x, y, 0, NULL, lpszString, cbString, NULL, MWTF_UC16 );
}

/* ascii or utf8 */
BOOL WINAPI
ExtTextOut(HDC hdc, int x, int y, UINT fuOptions, CONST RECT *lprc,
	LPCSTR lpszString, UINT cbCount, CONST INT *lpDx)
{
	return MwExtTextOut(hdc, x, y, fuOptions, lprc, lpszString,
		cbCount, lpDx, mwTextCoding);
}

/* unicode*/
BOOL WINAPI
ExtTextOutW(HDC hdc, int x, int y, UINT fuOptions, CONST RECT *lprc,
	LPCWSTR lpszString, UINT cbCount, CONST INT *lpDx)
{
	return MwExtTextOut(hdc, x, y, fuOptions, lprc, lpszString,
		cbCount, lpDx, MWTF_UC16);
}

/* internal version of ExtTextOut, passed flags for text data type*/
static BOOL
MwExtTextOut(HDC hdc, int x, int y, UINT fuOptions, CONST RECT *lprc,
	LPCVOID lpszString, UINT cbCount, CONST INT *lpDx, int flags)
{
	HWND	hwnd;
	POINT	pt;
	RECT	rc;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	pt.x = x;
	pt.y = y;
	if(MwIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	/* optionally fill passed rectangle*/
	if(lprc && (fuOptions&ETO_OPAQUE)) {
		rc = *lprc;
		if(MwIsClientDC(hdc))
			MapWindowPoints(hwnd, NULL, (LPPOINT)&rc, 2);

		/* fill rectangle with current background color*/
		GdSetForegroundColor(hdc->psd, hdc->bkcolor);
		GdFillRect(hdc->psd, rc.left, rc.top, rc.right - rc.left,
			rc.bottom - rc.top);
		GdSetUseBackground(FALSE);
	} else {
		/* use current background mode for text background draw*/
		GdSetUseBackground(hdc->bkmode == OPAQUE? TRUE: FALSE);
		/* always set background color in case GdArea is
		 * used to draw, which compares gr_foreground != gr_background
		 * if gr_usebg is false...
		 */
		/*if(hdc->bkmode == OPAQUE)*/
			GdSetBackgroundColor(hdc->psd, hdc->bkcolor);
	}

	if (cbCount == 0) {
		/* Special case - no text.  Used to fill rectangle. */
		return TRUE;
	}

	/* nyi: lpDx*/

	/* draw text in current text foreground and background color*/
	GdSetForegroundColor(hdc->psd, hdc->textcolor);
	//GdSetFont(hdc->font->pfont);

	/* this whole text alignment thing needs rewriting*/
	if((hdc->textalign & TA_BASELINE) == TA_BASELINE) {
		 /* this is not right... changed for kaffe port
		flags |= MWTF_TOP;
		 */
		flags |= MWTF_BASELINE;
	} else if(hdc->textalign & TA_BOTTOM) {
		MWCOORD	ph, pw, pb;

		if(lprc)
			pt.y += lprc->bottom - lprc->top;
		else {
			GdGetTextSize(hdc->font->pfont, lpszString, cbCount, &pw, &ph, &pb, flags);
			pt.y += ph;
		}
		flags |= MWTF_BOTTOM;
	} else
		flags |= MWTF_TOP;

	if((hdc->textalign & TA_CENTER) == TA_CENTER) {
		MWCOORD     ph, pw, pb;

		GdGetTextSize(hdc->font->pfont, lpszString, cbCount, &pw, &ph, &pb, flags);
		pt.x -= pw/2;
	} else if(hdc->textalign & TA_RIGHT) {
		MWCOORD     ph, pw, pb;

		GdGetTextSize(hdc->font->pfont, lpszString, cbCount, &pw, &ph, &pb, flags);
		pt.x -= pw;
	}
	GdText(hdc->psd, hdc->font->pfont, pt.x, pt.y, lpszString, cbCount, flags);

	return TRUE;
}

/* ascii or utf8 */
int WINAPI
DrawTextA(HDC hdc, LPCSTR lpString, int nCount, LPRECT lpRect, UINT uFormat)
{
	return MwDrawText(hdc, lpString, nCount, lpRect, uFormat, mwTextCoding);
}
#endif
/* unicode*/
int WINAPI
DrawText (HDC hdc, LPCWSTR lpString, int nCount, LPRECT lpRect, UINT uFormat)
{
//	return MwDrawText(hdc, lpString, nCount, lpRect, uFormat, MWTF_UC16);
	assert (0);
	return 0;
}

#if 0
#if OLD_DRAWTEXT
/*
 *  Check in text for the '&' chr, remove it from text and sets rect for pos
 */
static LPCSTR
MwCheckUnderlineChar(HDC hdc, char *text, int *pLen, LPRECT rcLine)
{
	int i;
	int txtlen;

	if (pLen)
		txtlen = *pLen;
	else
		txtlen = strlen(text);

	for (i = 0; i < txtlen; i++) {
		if ((text[i] == '&') && (i + 1 < txtlen)
		    && (text[i + 1] != '&')) {
			SIZE sz;
			TEXTMETRIC tm;

			GetTextMetrics(hdc, &tm);
			GetTextExtentPoint(hdc, text, i, &sz);
			rcLine->left = sz.cx;
			rcLine->top = 0;
			rcLine->bottom = tm.tmAscent + 1;
			GetTextExtentPoint(hdc, text + i + 1, 1, &sz);
			rcLine->right = rcLine->left + sz.cx;
			text = strdup(text);
			memmove(text + i, text + i + 1, txtlen - i);
			txtlen--;
			if (pLen)
				*pLen = txtlen;
			return text;
		}
	}
	return FALSE;
}

/* note: many DT_x aren't implemented in this function*/
/* internal version of DrawText, passed flags for text data type*/
static int
MwDrawText(HDC hdc, LPCVOID lpString, int nCount, LPRECT lpRect, UINT uFormat, int flags)
{
	MWCOORD	x, y, width, height, baseline;
	LPCSTR lpStrDup = NULL;
	RECT rcUline;

	if(!hdc)
		return 0;

	if(nCount == -1)
		nCount = strlen(lpString);

	if (!(uFormat & DT_NOPREFIX) && (lpStrDup = MwCheckUnderlineChar(hdc, lpString, &nCount, &rcUline)))
		lpString = lpStrDup;
	if(uFormat & (DT_CALCRECT|DT_CENTER|DT_RIGHT)) {
		GdGetTextSize(hdc->font->pfont, lpString, nCount,
			&width, &height, &baseline, flags);
	}
	x = lpRect->left;
	y = lpRect->top;

	if(uFormat & DT_CALCRECT) {
#if 0
		if (!(uFormat & DT_SINGLELINE)) {
			char *p=(char*)lpString, *pLast=p, *pMax=p;
			int iLenCur, iLenMax = 0, iLines = 0;
			long lMaxWidth = lpRect->right - lpRect->left;

			/* Determine line with maximum width */
			for (pLast=p, pMax=p; ; p++) {
				if (*p=='\n' || !*p) {
					if (lMaxWidth) {
						do {
							iLenCur = p - pLast;
							GdGetTextSize(hdc->font->pfont, pLast, iLenCur,
								&width, &height, &baseline, flags);
							if (width <= lMaxWidth) break;
							while (p>(char*)lpString && (*p==' ' || *p=='\t')) p--;
							while (p>(char*)lpString && !(*p==' ' || *p=='\t')) p--;

						} while (width > lMaxWidth && p>(char*)lpString);
					}
					if ((iLenCur = p - pLast) > iLenMax) {
						iLenMax = iLenCur;
						pMax = pLast;
					}
					pLast = p+1;
					iLines++;
				}
				if (!*p) break;
			}
			GdGetTextSize(hdc->font->pfont, pMax, iLenMax,
				&width, &height, &baseline, flags);
			height *= iLines;
		}
#endif
		lpRect->right = x + width;
		lpRect->bottom = y + height;
		if (lpStrDup)
			free(lpStrDup);
		return height;
	}

	if(uFormat & DT_CENTER)
		x = (lpRect->left + lpRect->right - width) / 2;
	else if(uFormat & DT_RIGHT)
		x = lpRect->right - width;

	if(uFormat & DT_VCENTER)
	    y = lpRect->bottom / 2 - (height / 2);

	if(uFormat & DT_BOTTOM) {
	    flags |= MWTF_BOTTOM;
	    y = lpRect->bottom;
	}

	/* draw text at DT_TOP using current fg, bg and bkmode*/
	MwExtTextOut(hdc, x, y, 0, NULL, lpString, nCount, NULL, flags);
	if (lpStrDup) {
		OffsetRect(&rcUline, x, y);
		SelectObject(hdc, GetStockObject(BLACK_PEN));
		MoveToEx(hdc, rcUline.left, rcUline.bottom, NULL);
		LineTo(hdc, rcUline.right, rcUline.bottom);
		free(lpStrDup);
	}
	return height;
}
#else
/*
 * MwDrawText Support routine: given a position, a string and length,
 *	process underline, ampersands and tabs, and output
 *	using mwExtTextOut ormwTabbedTextOut
 */
static void
mwDrawTextOut(HDC hDC, int x, int y, LPSTR str, int len, UINT uFormat, int flags)
{
	int i, j, k, dx, dy, x1;
	MWCOORD baseline;

	if(!hDC)
		return;

	for (i = 0, j = 0, k = 0, x1 = 0; i < len; i++) {
		if (str[i] == '&') {
			if (i) {
				if (uFormat & DT_EXPANDTABS) {
					mwTabbedTextOut(hDC, x, y, &str[j], i, 0, 0, x, FALSE, flags);
					x += LOWORD(mwTabbedTextOut(hDC, x, y, &str[j], i, 0, 0, 0, TRUE, flags));
				} else {
					MwExtTextOut(hDC, x, y, 0, NULL, &str[j], i - j, NULL, flags);
					GdGetTextSize(hDC->font->pfont, &str[j], i - j, &dx, &dy, &baseline, flags);
					x += dx;
				}
			}
			if (str[i + 1] == '&') {
				TextOut(hDC, x, y, "&", 1);
				GdGetTextSize(hDC->font->pfont, "&", 1, &dx, &dy, &baseline, flags);
				x += dx;
				i++;
			} else {
				x1 = x;
				k = i + 1;
			}
			j = i + 1;
		}
	}

	if (i > j) {
		if (uFormat & DT_EXPANDTABS)
			mwTabbedTextOut(hDC, x, y, &str[j], i - j, 0, 0, x, FALSE, flags);
		else
			MwExtTextOut(hDC, x, y, 0, NULL, &str[j], i - j, NULL, flags);
		if (k && str[k] != ' ') {
			GdGetTextSize(hDC->font->pfont, &str[k], 1, &dx, &dy, &baseline, flags);
			dy -= dy / 18 + 1;	/* underline position */
			MoveToEx(hDC, x1, y + dy, NULL);
			LineTo(hDC, x1 + dx, y + dy);
		}
	}
}

/*
 * Complex text output, supports:
 *	DT_BOTTOM, DT_CALCRECT, DT_CENTER, DT_EXTERNALLEADING,
 *	DT_LEFT, DT_RIGHT, DT_SINGLELINE, DT_TOP (default), DT_VCENTER,
 *	DT_WORDBREAK, DT_NOCLIP (default), DT_NOPREFIX, DT_EXPANDTABS
 */
static int WINAPI
MwDrawText(HDC hDC, LPCVOID lpsz, int cb, LPRECT lprc, UINT uFormat, int flags)
{
	LPSTR str = lpsz;
	int lineheight;
	int textwidth;
	int x, y, baseline, baselinefnt;
	int dx, dy;
	int nlines = 0;
	int tabsize = 8;
	int rtotal = 0;
	HPEN hPen = 0;
	TEXTMETRIC TextMetrics;
	int charwidth[256];

	if (!lprc || !hDC)
		return 0;

	if (uFormat & DT_TABSTOP) {
		tabsize = uFormat >> 8;		/* FIXME: not used*/
		uFormat &= 0xff;
	}

	GetCharWidth(hDC, 0, 255, charwidth);
	GetTextMetrics(hDC, (LPTEXTMETRIC)&TextMetrics);
	lineheight = TextMetrics.tmHeight;

	if (uFormat & DT_EXTERNALLEADING)
		lineheight += TextMetrics.tmExternalLeading;

	if (uFormat & DT_CALCRECT) {
		lprc->left = HIWORD(lprc->left) ? 0 : lprc->left;
		lprc->right = HIWORD(lprc->right) ? 0 : lprc->right;
		lprc->top = HIWORD(lprc->top) ? 0 : lprc->top;
		lprc->bottom = HIWORD(lprc->bottom) ? 0 : lprc->bottom;
	} else {
		hPen = CreatePen(PS_SOLID, 1, GetTextColor(hDC));	/* underline text color */
		hPen = SelectObject(hDC, hPen);
	}

	textwidth = lprc->right - lprc->left;
	baseline = lprc->top;

	if (cb == -1)
		cb = str? strlen(str): 0;

	while (cb) {
		int space = 0;
		int Ampersands = 0;
		int AmpersandPos = 0;
		int lbreak = 0;
		int place = 0;
		int cnt;

		rtotal = 0;

		for (cnt = 0; cnt < cb; cnt += MW_CHRNBYTE(str[cnt])) {
			switch (str[cnt]) {
			case 0:
				lbreak++;
				break;
			case ' ':
				space++;
				place = cnt;
				rtotal += charwidth[(BYTE) str[cnt]];
				break;
			case '\t':
				space++;
				place = cnt;
				break;
			case '&':
				if (uFormat & DT_NOPREFIX)
					rtotal += charwidth[(BYTE) str[cnt]];
				else {
					if (Ampersands && AmpersandPos + 1 == cnt) {
						rtotal += charwidth[(BYTE) str[cnt]];
					} else {
						AmpersandPos = cnt;
						Ampersands = 1;
					}
				}
				break;
			case '\r':
			case '\n':
				lbreak++;
				break;
			default:
				rtotal += charwidth[(BYTE) str[cnt]];
				break;
			}

			if (lbreak)
				break;

			/* check width */
			if (rtotal > textwidth) {
				if ((uFormat & DT_CALCRECT) && (uFormat & DT_SINGLELINE)) {
					lprc->right += (rtotal - textwidth);
					textwidth = lprc->right - lprc->left;
					continue;
				}
				if (uFormat & DT_WORDBREAK) {
					if (space) {
						cnt = place;
						GdGetTextSize(hDC->font->pfont, str, cnt - 1, &dx, &dy, &baselinefnt, flags);
						rtotal = dx;
					}
					cnt++;
					break;
				}
			}
		}

		/* calc width */
		dx = rtotal;
		dy = lineheight;

		if (uFormat & DT_RIGHT)
			x = lprc->right - dx + 1;
		else if (uFormat & DT_CENTER)
			x = (int) (lprc->left + lprc->right - dx) / 2;
		else
			x = lprc->left;

		/* calc height */
		if (uFormat & DT_VCENTER) {
			GdGetTextSize(hDC->font->pfont, str, cnt, &dx, &dy, &baselinefnt, flags);
			y = (int) (baseline + lprc->bottom - dy) / 2;
		} else if (uFormat & DT_BOTTOM)
			y = lprc->bottom - dy;
		else
			y = baseline;

		if ((uFormat & DT_CALCRECT) == 0) {
			/* handle prefixes */
			if (Ampersands)
				mwDrawTextOut(hDC, x, y, str, cnt, uFormat, flags);
			else {
				if (uFormat & DT_EXPANDTABS) {
					mwTabbedTextOut(hDC, x, y, str, cnt, 0, 0, x, FALSE, flags);
				} else {
					if (uFormat & DT_NOCLIP)
						MwExtTextOut(hDC, x, y, 0, NULL, str, cnt, NULL, flags);
					else
						MwExtTextOut(hDC, x, y, ETO_CLIPPED, lprc, str, cnt, NULL, flags);
				}
			}
		}

		nlines++;
		baseline += lineheight;

		if (uFormat & DT_CALCRECT)
			lprc->bottom = baseline;

		if (uFormat & DT_SINGLELINE)
			break;

		if (lbreak) {
			if (str[cnt] == '\r')
				cnt++;
			if (str[cnt] == '\n')
				cnt++;
		}

		str += cnt;
		cb -= cnt;

		if (baseline > lprc->bottom) {
			break;
		}
	}

	if ((uFormat & DT_CALCRECT) && !(uFormat & DT_SINGLELINE) && nlines == 1)
		lprc->right = rtotal + lprc->left;

	if (hPen) {
		hPen = SelectObject(hDC, hPen);
		DeleteObject(hPen);
	}

	return baseline - lprc->top;
}
#endif /* !OLD_DRAWTEXT*/


/* Microwindows only*/
BOOL WINAPI
DrawDIB(HDC hdc,int x,int y,PMWIMAGEHDR pimage)
{
	HWND		hwnd;
	POINT		pt;

	hwnd = MwPrepareDC(hdc);
	if(!hwnd || !pimage)
		return FALSE;
	pt.x = x;
	pt.y = y;
	if(MwIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	GdDrawImage(hdc->psd, pt.x, pt.y, pimage);
	return TRUE;
}
#endif

/* define color scheme: A (tan), B (winstd) or C (old)*/
#if ELKS
#define B
#else
#define A
#endif

#define A_RGB(r,g,b)
#define B_RGB(r,g,b)
#define C_RGB(r,g,b)

#ifdef A
#undef  A_RGB
#define A_RGB(r,g,b)	RGB(r,g,b),
#endif
#ifdef B
#undef  B_RGB
#define B_RGB(r,g,b)	RGB(r,g,b),
#endif
#ifdef C
#undef  C_RGB
#define C_RGB(r,g,b)	RGB(r,g,b),
#endif

static COLORREF sysColors[MAXSYSCOLORS] = {
	RGB(192, 192, 192),	/* COLOR_SCROLLBAR          0*/
	RGB(  0, 128, 128),     /* COLOR_BACKGROUND          */
	A_RGB(128,   0,   0)	/* COLOR_ACTIVECAPTION       */
	B_RGB(128,   0, 128)	/* COLOR_ACTIVECAPTION       */
	C_RGB(128,   0, 128)	/* COLOR_ACTIVECAPTION       */
	A_RGB(162, 141, 104)	/* COLOR_INACTIVECAPTION     */
	B_RGB(128, 128, 128)  	/* COLOR_INACTIVECAPTION     */
	C_RGB(  0,  64, 128)	/* COLOR_INACTIVECAPTION     */
	RGB(192, 192, 192),     /* COLOR_MENU                */
	RGB(255, 255, 255),     /* COLOR_WINDOW             5*/
	RGB(  0,   0,   0),     /* COLOR_WINDOWFRAME         */
	RGB(  0,   0,   0),     /* COLOR_MENUTEXT            */
	RGB(  0,   0,   0),     /* COLOR_WINDOWTEXT          */
	RGB(255, 255, 255),     /* COLOR_CAPTIONTEXT         */
	RGB(192, 192, 192),   	/* COLOR_ACTIVEBORDER      10*/
	RGB(192, 192, 192), 	/* COLOR_INACTIVEBORDER      */
	RGB(128, 128, 128),   	/* COLOR_APPWORKSPACE        */
	RGB(128,   0,   0),     /* COLOR_HIGHLIGHT           */
	RGB(255, 255, 255),     /* COLOR_HIGHLIGHTTEXT       */
	A_RGB(213, 204, 187)	/* COLOR_BTNFACE           15*/
	B_RGB(192, 192, 192)	/* COLOR_BTNFACE           15*/
	C_RGB(160, 160, 160)	/* COLOR_BTNFACE           15*/
	A_RGB(162, 141, 104)	/* COLOR_BTNSHADOW           */
	B_RGB(128, 128, 128)	/* COLOR_BTNSHADOW           */
	C_RGB(128, 128, 128)	/* COLOR_BTNSHADOW           */
	RGB( 64,  64,  64),     /* COLOR_GRAYTEXT            */
	RGB(  0,   0,   0),     /* COLOR_BTNTEXT             */
	RGB(192, 192, 192),    	/* COLOR_INACTIVECAPTIONTEXT */
	A_RGB(234, 230, 221)  	/* COLOR_BTNHIGHLIGHT      20*/
	B_RGB(255, 255, 255)  	/* COLOR_BTNHIGHLIGHT      20*/
	C_RGB(223, 223, 223)  	/* COLOR_BTNHIGHLIGHT      20*/
	RGB(  0,   0,   0),     /* COLOR_3DDKSHADOW          */
	A_RGB(213, 204, 187)	/* COLOR_3DLIGHT             */
	B_RGB(223, 223, 223)	/* COLOR_3DLIGHT             */
	C_RGB(192, 192, 192)	/* COLOR_3DLIGHT             */
	RGB(  0,   0,   0),     /* COLOR_INFOTEXT            */
	RGB(225, 255, 255), 	/* COLOR_INFOBK              */
	RGB(184, 180, 184),  	/* COLOR_ALTERNATEBTNFACE  25*/
	RGB(  0,   0, 255),     /* COLOR_HOTLIGHT 	     */
	RGB( 16, 132, 208),  	/* COLOR_GRADIENTACTIVECAPTION */
	RGB(184, 180, 184) 	/* COLOR_GRADIENTINACTIVECAPTION 28*/
};

DWORD WINAPI
GetSysColor(int nIndex)
{
	if(nIndex >= 0 && nIndex < MAXSYSCOLORS)
		return sysColors[nIndex];
	return 0;
}

COLORREF WINAPI
SetSysColor(int nIndex, COLORREF crColor)	/* Microwindows only*/
{
	COLORREF oldColor;

	if(nIndex >= 0 && nIndex < MAXSYSCOLORS) {
		oldColor = sysColors[nIndex];
		sysColors[nIndex] = crColor;
		return oldColor;
	}
	return 0;
}

static MWBRUSHOBJ syscolorBrushes[MAXSYSCOLORS];

HBRUSH WINAPI
GetSysColorBrush(int nIndex)
{
	MWBRUSHOBJ *hbr;

	if(nIndex >= 0 && nIndex < MAXSYSCOLORS) {
		hbr = &syscolorBrushes[nIndex];
		hbr->hdr.type = OBJ_BRUSH;
		hbr->hdr.stockobj = TRUE;
		hbr->style = BS_SOLID;
		hbr->color = sysColors[nIndex];
		return (HBRUSH)hbr;
	}
	return 0;
}


static MWBRUSHOBJ OBJ_WHITE_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(255, 255, 255)
};

static MWBRUSHOBJ OBJ_LTGRAY_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(192, 192, 192)
};

static MWBRUSHOBJ OBJ_GRAY_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(128, 128, 128)
};

static MWBRUSHOBJ OBJ_DKGRAY_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(32, 32, 32)
};

static MWBRUSHOBJ OBJ_BLACK_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_SOLID, RGB(0, 0, 0)
};

static MWBRUSHOBJ OBJ_NULL_BRUSH = {
	{OBJ_BRUSH, TRUE}, BS_NULL, RGB(0, 0, 0)
};

static MWPENOBJ OBJ_WHITE_PEN = {
	{OBJ_PEN, TRUE}, PS_SOLID, RGB(255, 255, 255)
};

static MWPENOBJ OBJ_BLACK_PEN = {
	{OBJ_PEN, TRUE}, PS_SOLID, RGB(0, 0, 0)
};

static MWPENOBJ OBJ_NULL_PEN = {
	{OBJ_PEN, TRUE}, PS_NULL, RGB(0, 0, 0)
};
#if 0
static MWFONTOBJ OBJ_OEM_FIXED_FONT = {
	{OBJ_FONT, TRUE}, NULL, MWFONT_SYSTEM_FIXED	/* was MWFONT_OEM_FIXED*/
};

static MWFONTOBJ OBJ_ANSI_FIXED_FONT = {
	{OBJ_FONT, TRUE}, NULL, MWFONT_SYSTEM_FIXED
};

static MWFONTOBJ OBJ_ANSI_VAR_FONT = {
	{OBJ_FONT, TRUE}, NULL, MWFONT_SYSTEM_VAR
};

static MWFONTOBJ OBJ_SYSTEM_FONT = {
	{OBJ_FONT, TRUE}, NULL, MWFONT_SYSTEM_VAR
};

static MWFONTOBJ OBJ_DEVICE_DEFAULT_FONT = {
	{OBJ_FONT, TRUE}, NULL, MWFONT_SYSTEM_FIXED	/* was MWFONT_OEM_FIXED*/
};

static MWFONTOBJ OBJ_SYSTEM_FIXED_FONT = {
	{OBJ_FONT, TRUE}, NULL, MWFONT_SYSTEM_FIXED
};

static MWFONTOBJ OBJ_DEFAULT_GUI_FONT = {
	{OBJ_FONT, TRUE}, NULL, MWFONT_SYSTEM_VAR	/* was MWFONT_GUI_VAR*/
};
#else
static MWFONTOBJ OBJ_HII_FIXED_FONT = {
  {OBJ_FONT, TRUE}, L"", 19, 0
};
#endif

static MWBRUSHOBJ OBJ_DC_BRUSH = {
  {OBJ_BRUSH, TRUE}, BS_SOLID, RGB(255, 255 , 255)
};

static MWPENOBJ OBJ_DC_PEN = {
  {OBJ_PEN, TRUE}, PS_SOLID, RGB(255, 255, 255)
};

static struct hgdiobj *stockObjects[MAXSTOCKOBJECTS] = {
	(struct hgdiobj *)&OBJ_WHITE_BRUSH,		/* WHITE_BRUSH*/
	(struct hgdiobj *)&OBJ_LTGRAY_BRUSH,		/* LTGRAY_BRUSH*/
	(struct hgdiobj *)&OBJ_GRAY_BRUSH,		/* GRAY_BRUSH*/
	(struct hgdiobj *)&OBJ_DKGRAY_BRUSH,		/* DKGRAY_BRUSH*/
	(struct hgdiobj *)&OBJ_BLACK_BRUSH,		/* BLACK_BRUSH*/
	(struct hgdiobj *)&OBJ_NULL_BRUSH,		/* NULL_BRUSH*/
	(struct hgdiobj *)&OBJ_WHITE_PEN,		/* WHITE_PEN*/
	(struct hgdiobj *)&OBJ_BLACK_PEN,		/* BLACK_PEN*/
	(struct hgdiobj *)&OBJ_NULL_PEN,		/* NULL_PEN*/
	(struct hgdiobj *)NULL,
	(struct hgdiobj *)&OBJ_HII_FIXED_FONT,	/* OEM_FIXED_FONT*/
	(struct hgdiobj *)&OBJ_HII_FIXED_FONT,	/* ANSI_FIXED_FONT*/
	(struct hgdiobj *)&OBJ_HII_FIXED_FONT,	/* ANSI_VAR_FONT*/
	(struct hgdiobj *)&OBJ_HII_FIXED_FONT,	/* SYSTEM_FONT*/
	(struct hgdiobj *)&OBJ_HII_FIXED_FONT,	/* DEVICE_DEFAULT_FONT*/
	(struct hgdiobj *)NULL,				/* DEFAULT_PALETTE*/
	(struct hgdiobj *)&OBJ_HII_FIXED_FONT,	/* SYSTEM_FIXED_FONT*/
	(struct hgdiobj *)&OBJ_HII_FIXED_FONT,	/* DEFAULT_GUI_FONT*/
	(struct hgdiobj *)&OBJ_DC_BRUSH,		/* SYSTEM_FIXED_FONT*/
	(struct hgdiobj *)&OBJ_DC_PEN,			/* DEFAULT_GUI_FONT*/
};

HGDIOBJ WINAPI
GetStockObject(int nObject)
{
	HGDIOBJ		pObj;
	//MWFONTOBJ *	pFont;

	if(nObject >= 0 && nObject < MAXSTOCKOBJECTS) {
		pObj = stockObjects[nObject];

		//
		// only use hii fixed font
		//
		if(pObj->hdr.type == OBJ_FONT) {
			return pObj;
		}

		/* implement multiple color schemes with
		 * standard background brushes...
		 */
		switch(nObject) {
		case LTGRAY_BRUSH:
		case GRAY_BRUSH:
			((MWBRUSHOBJ *)pObj)->color = GetSysColor(COLOR_BTNFACE);
			break;
		case DKGRAY_BRUSH:
			((MWBRUSHOBJ *)pObj)->color = GetSysColor(COLOR_BTNSHADOW);
			break;
		}
		return pObj;
	}
	return NULL;
}

HGDIOBJ WINAPI
SelectObject(HDC hdc, HGDIOBJ hObject)
{
	HGDIOBJ		objOrg;
	MWBITMAPOBJ *	pb;

	if(!hdc || !hObject)
		return NULL;

	switch(hObject->hdr.type) {
	case OBJ_PEN:
		objOrg = (HGDIOBJ)hdc->pen;
		hdc->pen = (MWPENOBJ *)hObject;
		break;
	case OBJ_BRUSH:
		objOrg = (HGDIOBJ)hdc->brush;
		hdc->brush = (MWBRUSHOBJ *)hObject;
		break;
	case OBJ_FONT:
		objOrg = (HGDIOBJ)hdc->font;
		hdc->font = (MWFONTOBJ *)hObject;
		break;
	case OBJ_BITMAP:
		/* must be memory dc to select bitmap*/
		if(!(hdc->psd->flags&PSF_MEMORY))
			return NULL;
		objOrg = (HGDIOBJ)hdc->bitmap;

		/* setup mem dc for drawing into bitmap*/
		pb = (MWBITMAPOBJ *)hObject;

		/* init memory context*/
		if (!hdc->psd->MapMemGC(hdc->psd, pb->width, pb->height,
			pb->planes, pb->bpp, pb->data_format, pb->pitch, pb->size, &pb->bits[0]))
				return NULL;

		hdc->bitmap = (MWBITMAPOBJ *)hObject;
	    break;
#if UPDATEREGIONS
	case OBJ_REGION:
		/*objOrg = (HGDIOBJ)hdc->region;*/
		objOrg = NULL;	/* FIXME? hdc->region is destroyed below*/
		SelectClipRgn(hdc, (HRGN)hObject);
		break;
#endif
	default:
		return NULL;
	}

	return objOrg;
}

BOOL WINAPI
DeleteObject(HGDIOBJ hObject)
{
	if(!hObject || hObject->hdr.stockobj)
		return FALSE;
	//if(hObject->hdr.type == OBJ_FONT)
	//	GdDestroyFont(((MWFONTOBJ *)hObject)->pfont);
	if(hObject->hdr.type == OBJ_REGION)
		GdDestroyRegion(((MWRGNOBJ *)hObject)->rgn);
	GdItemFree(hObject);
	return TRUE;
}

#if UPDATEREGIONS
/* region is passed in client coords (win32 api doc bug)*/
int WINAPI
SelectClipRgn(HDC hdc, HRGN hrgn)
{
	return ExtSelectClipRgn(hdc, hrgn, RGN_COPY);
}

/*
 * Select a user clip region into DC, recalculate final clipregion.
 * Only a copy of the passed region is used.
 */
/* region is passed in client coords (win32 api doc bug)*/
int WINAPI
ExtSelectClipRgn(HDC hdc, HRGN hrgn, int fnMode)
{
	HRGN	newrgn;

	if(!hdc)
		return ERROR;
	if(hdc->region != (MWRGNOBJ *)hrgn) {
		/* combine region if not null*/
		if(hrgn) {
			newrgn = CreateRectRgn(0, 0, 0, 0);

			/*
			 * Temporarily convert region from
			 * client coords to screen coords, since
			 * hwnd->update is kept in screen coords.
			 */
			if (hdc->flags & PSF_MEMORY) {

				if(fnMode == RGN_COPY)
					CombineRgn(newrgn, hrgn, NULL, fnMode);
				else CombineRgn(newrgn, (HRGN)hdc->region, hrgn,fnMode);


			} else {

				OffsetRgn(hrgn, hdc->hwnd->clirect.left,
					hdc->hwnd->clirect.top);

				if(fnMode == RGN_COPY)
					CombineRgn(newrgn, hrgn, NULL, fnMode);
				else CombineRgn(newrgn, (HRGN)hdc->region, hrgn,fnMode);

				/* convert passed region back to client coords*/
				OffsetRgn(hrgn, -hdc->hwnd->clirect.left,
					-hdc->hwnd->clirect.top);
			}


			hrgn = newrgn;
		}
		DeleteObject((HRGN)hdc->region);
		hdc->region = (MWRGNOBJ *)hrgn;

		/* force recalc of clipregion*/
		cliphdc = NULL;
		MwPrepareDC(hdc);
	}
	if(hrgn)
		return ((MWRGNOBJ *)hrgn)->rgn->type;
	return NULLREGION;
}

/* update region is returned in client coordinates*/
int WINAPI
GetUpdateRgn(HWND hwnd, HRGN hrgn, BOOL bErase)
{
	/* FIXME bErase*/
	if(!hwnd)
		return ERROR;

	/* convert internal update region to client coords*/
	GdOffsetRegion(hwnd->update, -hwnd->clirect.left, -hwnd->clirect.top);
	GdCopyRegion(((MWRGNOBJ *)hrgn)->rgn, hwnd->update);
	GdOffsetRegion(hwnd->update, hwnd->clirect.left, hwnd->clirect.top);
	return hwnd->update->type;
}
#endif /* UPDATEREGIONS*/

/* update rectangle is returned in client coords*/
BOOL WINAPI
GetUpdateRect(HWND hwnd, LPRECT lpRect, BOOL bErase)
{
	/* FIXME bErase*/
	if(!hwnd)
		return FALSE;
#if UPDATEREGIONS
	if(lpRect) {
		*lpRect = hwnd->update->extents;
		/* convert to client coords*/
		ScreenToClient(hwnd, (LPPOINT)&lpRect->left);
		ScreenToClient(hwnd, (LPPOINT)&lpRect->right);
	}

	/* return TRUE if update region is non-empty*/
	return hwnd->update->type != NULLREGION;
#else
	GetClientRect(hwnd, lpRect);
	return TRUE;
#endif
}

HBRUSH WINAPI
CreateSolidBrush(COLORREF crColor)
{
	MWBRUSHOBJ *hbr;

	hbr = GdItemNew(MWBRUSHOBJ);
	if(!hbr)
		return NULL;
	hbr->hdr.type = OBJ_BRUSH;
	hbr->hdr.stockobj = FALSE;
	hbr->style = BS_SOLID;
	hbr->color = crColor;
	return (HBRUSH)hbr;
}

HPEN WINAPI
CreatePen(int nPenStyle, int nWidth, COLORREF crColor)
{
	MWPENOBJ *hpen;

	/* fix: nWidth > 1*/
	hpen = GdItemNew(MWPENOBJ);
	if(!hpen)
		return NULL;
	hpen->hdr.type = OBJ_PEN;
	hpen->hdr.stockobj = FALSE;
	hpen->style = nPenStyle;
	hpen->color = crColor;
	return (HPEN)hpen;
}

HBITMAP WINAPI
CreateCompatibleBitmap(HDC hdc, int nWidth, int nHeight)
{
	MWBITMAPOBJ *	hbitmap;
	unsigned int size;
	unsigned int pitch;

	if(!hdc)
		return NULL;

	nWidth = MWMAX(nWidth, 1);
	nHeight = MWMAX(nHeight, 1);

	/* calc memory allocation size and pitch from width and height*/
	if(!GdCalcMemGCAlloc(hdc->psd, nWidth, nHeight, 0, 0, &size, &pitch))
		return NULL;

	/* allocate gdi object*/
	hbitmap = (MWBITMAPOBJ *)GdItemAlloc(sizeof(MWBITMAPOBJ)-1+size);
	if(!hbitmap)
		return NULL;
	hbitmap->hdr.type = OBJ_BITMAP;
	hbitmap->hdr.stockobj = FALSE;
	hbitmap->width = nWidth;
	hbitmap->height = nHeight;

	/* create compatible with hdc*/
	hbitmap->planes = hdc->psd->planes;
	hbitmap->bpp = hdc->psd->bpp;
	hbitmap->data_format = hdc->psd->data_format;
	hbitmap->pitch = pitch;
	hbitmap->size = size;

	return (HBRUSH)hbitmap;
}

HBITMAP CreateDIBSection(
  HDC hdc, CONST BITMAPINFO *pbmi, UINT iUsage,
  VOID **ppvBits, HANDLE hSection, DWORD dwOffset)
{
	MWBITMAPOBJ *	hbitmap;
	unsigned int size;
	unsigned int pitch;
	PSD psd = hdc? hdc->psd: &scrdev;

	/* calc memory allocation size and pitch from width and height*/
	if(!GdCalcMemGCAlloc(psd, pbmi->bmiHeader.biWidth, pbmi->bmiHeader.biHeight,
		pbmi->bmiHeader.biPlanes, pbmi->bmiHeader.biBitCount, &size, &pitch))
			return NULL;

	/* allocate gdi object*/
	hbitmap = (MWBITMAPOBJ *)GdItemAlloc(sizeof(MWBITMAPOBJ)-1+size);
	if(!hbitmap)
		return NULL;
	hbitmap->hdr.type = OBJ_BITMAP;
	hbitmap->hdr.stockobj = FALSE;
	hbitmap->width = pbmi->bmiHeader.biWidth;
	hbitmap->height = pbmi->bmiHeader.biHeight;
	hbitmap->planes = pbmi->bmiHeader.biPlanes;
	hbitmap->bpp = pbmi->bmiHeader.biBitCount;
	hbitmap->pitch = pitch;
	hbitmap->size = size;

	if (ppvBits) *ppvBits = &hbitmap->bits[0];

	return (HBITMAP)hbitmap;
}

/* return NULL if no driver bitblit available*/
HDC WINAPI
CreateCompatibleDC(HDC hdc)
{
	HDC	hdcmem;
	PSD	psd;
	PSD	mempsd;

	/* allow NULL hdc to mean screen*/
	psd = hdc? hdc->psd: &scrdev;

	/* allocate memory device, if driver doesn't blit will fail*/
	mempsd = psd->AllocateMemGC(psd);
	if(!mempsd)
		return NULL;

	/* allocate a DC for DesktopWindow*/
	hdcmem = GetDCEx(hdc ? hdc->hwnd : NULL, NULL, DCX_DEFAULTCLIP);
	if(!hdcmem) {
		mempsd->FreeMemGC(mempsd);
		return NULL;
	}
	hdcmem->psd = mempsd;

	hdcmem->flags |= DCX_WINDOW;
	/* select in default bitmap to setup mem device parms*/
	SelectObject(hdcmem, (HGDIOBJ)&default_bitmap);
	return hdcmem;
}

BOOL WINAPI
BitBlt(HDC hdcDest, int nXDest, int nYDest, int nWidth, int nHeight,
	HDC hdcSrc, int nXSrc, int nYSrc, DWORD dwRop)
{
	/* use stretch blit with equal src and dest width/height*/
	return StretchBlt(hdcDest, nXDest, nYDest, nWidth, nHeight,
		hdcSrc, nXSrc, nYSrc, nWidth, nHeight, dwRop);
}


BOOL WINAPI
StretchBlt(HDC hdcDest, int nXOriginDest, int nYOriginDest, int nWidthDest,
	int nHeightDest, HDC hdcSrc, int nXOriginSrc, int nYOriginSrc,
	int nWidthSrc, int nHeightSrc, DWORD dwRop)
{

	HWND	hwnd;
	POINT	dst, src;

	if(!hdcDest || !hdcSrc)
		return FALSE;
	dst.x = nXOriginDest;
	dst.y = nYOriginDest;
	src.x = nXOriginSrc;
	src.y = nYOriginSrc;

	/* if src screen DC, convert coords*/
	/* FIXME: src clipping isn't checked, only one set of cliprects also*/

	//
	// DestDc must set clipping for memory or screen
	//
	if ((hwnd = MwPrepareDC (hdcDest)) == NULL) {
		return FALSE;
	}
	if(!MwIsMemDC(hdcSrc) && MwIsClientDC(hdcSrc)) {
		ClientToScreen(hwnd, &src);
	}

	/* if dst screen DC, convert coords and set clipping*/
	/* FIXME: if dest is also screen, src clipping will be overwritten*/
	if(!MwIsMemDC(hdcDest) && MwIsClientDC(hdcDest)) {
		ClientToScreen(hwnd, &dst);
	}

	if (nWidthDest == nWidthSrc && nHeightDest == nHeightSrc) {
		GdBlit(hdcDest->psd, dst.x, dst.y, nWidthDest, nHeightDest,
			hdcSrc->psd, src.x, src.y, dwRop);
	} else {
		GdStretchBlit(hdcDest->psd, dst.x, dst.y,
			dst.x + nWidthDest, dst.y + nHeightDest,
			hdcSrc->psd, src.x, src.y,
			src.x + nWidthSrc, src.y + nHeightSrc, dwRop);
	}

	return TRUE;
}
#if 0
UINT WINAPI
GetSystemPaletteEntries(HDC hdc,UINT iStartIndex,UINT nEntries,
	LPPALETTEENTRY lppe)
{
	UINT		i;
	MWPALENTRY	rgb;

	/* currently, we only work for screen device*/
	if(!hdc || hdc->psd != &scrdev)
		return 0;

	for(i=0; i<nEntries; ++i) {
		if(!GdGetPalette(hdc->psd, i+iStartIndex, 1, &rgb))
			break;
		lppe->peRed = rgb.r;
		lppe->peGreen = rgb.g;
		lppe->peBlue = rgb.b;
		lppe->peFlags = 0;
		++lppe;
	}
	return i;
}
#endif

/* allow NULL hdc for scrdev*/
int WINAPI
GetDeviceCaps(HDC hdc, int nIndex)
{
	PSD	psd;

	if (!hdc)
		psd = &scrdev;
	else psd = hdc->psd;

	switch(nIndex) {
	case HORZRES:
		return psd->xvirtres;
	case VERTRES:
		return psd->yvirtres;
	case BITSPIXEL:
		return psd->bpp;
	case PLANES:
		return psd->planes;
	case LOGPIXELSX:
	case LOGPIXELSY:
		return 96;
	case SIZEPALETTE:
		if (psd->bpp <= 8)
			return psd->ncolors;
		break;
	}
	return 0;
}

#if 0
/*
 *  Draw a rectangle indicating focus
 */
BOOL WINAPI
DrawFocusRect(HDC hdc, LPRECT prect)
{
	uint32_t dm = 0xAAAAAAAA;
	int dc = 32;
	int oldmode = GdSetMode(MWROP_XOR);
	HPEN holdpen = SelectObject(hdc,
		CreatePen(PS_SOLID, 1, RGB(255, 255, 255)));

	GdSetDash(&dm, &dc);
	SelectObject(hdc, GetStockObject(NULL_BRUSH));
	Rectangle(hdc, prect->left - 1, prect->top, prect->right + 1,
		  prect->bottom);
	GdSetDash(&dm, &dc);
	DeleteObject(SelectObject(hdc, holdpen));
	GdSetMode(oldmode);
	return TRUE;
}


/* ascii or utf8 */
static LONG
mwTabbedTextOut(HDC hdc, int x, int y, LPCTSTR lpszString, int cbString,
	int ntabs, LPINT lpTabStops, int nTabOrigin, BOOL noDraw, int flags)
{
	TEXTMETRIC tm;
	int count;
	int tot;
	int xw, xh, xb;
	int tabPos = x;
	int deftab = 32;
	int orgx = x;
	LPTSTR szShaped;
	LPCTSTR pstr;
	LPINT pTab = lpTabStops;
	unsigned long attrib = 0;

	if (!hdc)
		return 0;

	if (GetTextMetrics(hdc, &tm))
		deftab = 8 * tm.tmAveCharWidth;

	if (pTab == NULL)
		ntabs = 0;

	if (ntabs == 1) {
		deftab = *pTab;
		ntabs = 0;
	}

	if (cbString == -1)
		cbString = strlen(lpszString);

	/* Duplicate text. If coding is UTF-8, generate it by checking shape/joining*/
	if (flags & MWTF_UTF8)
		szShaped = doCharShape_UTF8(lpszString, cbString, &cbString, &attrib);
	else
		szShaped = strdup(lpszString);

	if (szShaped == NULL)
		return 0;

	pstr = szShaped;
	tot = cbString;
	//GdSetFont(hdc->font->pfont);
	while (tot > 0) {
		for (count = 0; (count < tot) && (pstr[count] != '\t'); count++)
		     continue;

		GdGetTextSize(hdc->font->pfont, pstr, count, &xw, &xh, &xb, flags);

		while ((ntabs > 0) && (nTabOrigin + *pTab <= (x + xw))) {
			pTab++;
			ntabs--;
		}

		if (count == tot)
			tabPos = x + xw;
		else if (ntabs > 0)
			tabPos = nTabOrigin + *pTab;
		else
			tabPos = nTabOrigin + ((x + xw - nTabOrigin) / deftab + 1) * deftab;

		if (!noDraw) {
			RECT rect;

			rect.left = x;
			rect.top = y;
			rect.right = x + tabPos;
			rect.bottom = y + xh;
			//TextOut ( hdc, x, y, pstr, count );
			if (attrib & TEXTIP_EXTENDED) {
				LPSTR virtText = doCharBidi_UTF8(pstr, count, NULL, NULL, &attrib);
				if (virtText) {
					if (attrib & TEXTIP_RTOL)
						DrawTextA(hdc, virtText, count, &rect, DT_RIGHT | DT_SINGLELINE | DT_TOP);
					else
						DrawTextA(hdc, virtText, count, &rect, DT_LEFT | DT_SINGLELINE | DT_TOP);
					free(virtText);
				}
			} else
				DrawTextA(hdc, pstr, count, &rect, DT_LEFT | DT_SINGLELINE | DT_TOP);
		}

		x = tabPos;
		tot -= (count + 1);
		pstr += (count + 1);
	}

	free(szShaped);
	return MAKELONG((tabPos - orgx), xh);
}


/* ascii or utf8 */
LONG WINAPI
TabbedTextOut(HDC hdc, int x, int y, LPCTSTR lpszString, int cbString,
	int ntabs, LPINT lpTabStops, int nTabOrigin)
{
	return mwTabbedTextOut(hdc, x, y, lpszString, cbString, ntabs,
			       lpTabStops, nTabOrigin, FALSE, mwTextCoding);
}


DWORD WINAPI
GetTabbedTextExtent(HDC hdc, int x, int y, LPCTSTR lpszString, int cbString,
	int ntabs, LPINT lpTabStops)
{
	return mwTabbedTextOut(hdc, x, y, lpszString, cbString, ntabs,
			       lpTabStops, 0, TRUE, mwTextCoding);
}
#endif

int WINAPI
MulDiv(int nMultiplicand, int nMultiplier, int nDivisor)
{
    int         nResult;
    long        lMulti;
    lMulti = nMultiplicand * nMultiplier;
    if ( lMulti < 0 )
        lMulti -= nDivisor >> 1;
    else
        lMulti += nDivisor >> 1;
    if(nDivisor == 0) nDivisor = 1;
    nResult = lMulti / nDivisor;
    return nResult;
}

COLORREF
SetDCBrushColor (
  HDC Hdc,
  COLORREF Color
  )
{

  COLORREF OldColor;

  OldColor = OBJ_DC_BRUSH.color;
  OBJ_DC_BRUSH.color = Color;

  return OldColor;
}

COLORREF
SetDCPenColor (
  HDC Hdc,
  COLORREF Color
  )
{

  COLORREF OldColor;

  OldColor = OBJ_DC_PEN.color;
  OBJ_DC_PEN.color = Color;

  return OldColor;
}


typedef struct {
  CHAR16  WChar;
  UINT32  Width;
} UNICODE_WIDTH_ENTRY;


UNICODE_WIDTH_ENTRY mFBUnicodeWidthTable[] = {
  //
  // General script area
  //
  {(CHAR16)0x1FFF,  1},
  /*
   * Merge the blocks and replace them with the above entry as they fall to
   * the same category and they are all narrow glyph. This will reduce search
   * time and table size. The merge will omit the reserved code.
   *
   * Remove the above item if below is un-commented.
   *
  {(CHAR16)0x007F,  1},       // C0 controls and basic Latin. 0x0000-0x007F
  {(CHAR16)0x00FF,  1},       // C1 controls and Latin-1 support. 0x0080-0x00FF
  {(CHAR16)0x017F,  1},       // Latin extended-A. 0x0100-0x017F
  {(CHAR16)0x024F,  1},       // Latin extended-B. 0x0180-0x024F
  {(CHAR16)0x02AF,  1},       // IPA extensions. 0x0250-0x02AF
  {(CHAR16)0x02FF,  1},       // Spacing modifier letters. 0x02B0-0x02FF
  {(CHAR16)0x036F,  1},       // Combining diacritical marks. 0x0300-0x036F
  {(CHAR16)0x03FF,  1},       // Greek. 0x0370-0x03FF
  {(CHAR16)0x04FF,  1},       // Cyrillic. 0x0400-0x04FF
  {(CHAR16)0x052F,  0},       // Unassigned. As Armenian in ver3.0. 0x0500-0x052F
  {(CHAR16)0x058F,  1},       // Armenian. 0x0530-0x058F
  {(CHAR16)0x05FF,  1},       // Hebrew. 0x0590-0x05FF
  {(CHAR16)0x06FF,  1},       // Arabic. 0x0600-0x06FF
  {(CHAR16)0x08FF,  0},       // Unassigned. 0x0700-0x08FF
  {(CHAR16)0x097F,  1},       // Devanagari. 0x0900-0x097F
  {(CHAR16)0x09FF,  1},       // Bengali. 0x0980-0x09FF
  {(CHAR16)0x0A7F,  1},       // Gurmukhi. 0x0A00-0x0A7F
  {(CHAR16)0x0AFF,  1},       // Gujarati. 0x0A80-0x0AFF
  {(CHAR16)0x0B7F,  1},       // Oriya. 0x0B00-0x0B7F
  {(CHAR16)0x0BFF,  1},       // Tamil. (See page 7-92). 0x0B80-0x0BFF
  {(CHAR16)0x0C7F,  1},       // Telugu. 0x0C00-0x0C7F
  {(CHAR16)0x0CFF,  1},       // Kannada. (See page 7-100). 0x0C80-0x0CFF
  {(CHAR16)0x0D7F,  1},       // Malayalam (See page 7-104). 0x0D00-0x0D7F
  {(CHAR16)0x0DFF,  0},       // Unassigned. 0x0D80-0x0DFF
  {(CHAR16)0x0E7F,  1},       // Thai. 0x0E00-0x0E7F
  {(CHAR16)0x0EFF,  1},       // Lao. 0x0E80-0x0EFF
  {(CHAR16)0x0FBF,  1},       // Tibetan. 0x0F00-0x0FBF
  {(CHAR16)0x109F,  0},       // Unassigned. 0x0FC0-0x109F
  {(CHAR16)0x10FF,  1},       // Georgian. 0x10A0-0x10FF
  {(CHAR16)0x11FF,  1},       // Hangul Jamo. 0x1100-0x11FF
  {(CHAR16)0x1DFF,  0},       // Unassigned. 0x1200-0x1DFF
  {(CHAR16)0x1EFF,  1},       // Latin extended additional. 0x1E00-0x1EFF
  {(CHAR16)0x1FFF,  1},       // Greek extended. 0x1F00-0x1FFF
  *
  */

  //
  // Symbol area
  //
  {(CHAR16)0x2FFF,  1},
  /*
   * Merge the blocks and replace them with the above entry as they fall to
   * the same category and they are all narrow glyph. This will reduce search
   * time and table size. The merge will omit the reserved code.
   *
   * Remove the above item if below is un-commented.
   *
  {(CHAR16)0x206F,  1},       // General punctuation. (See page7-154). 0x200-0x206F
  {(CHAR16)0x209F,  1},       // Superscripts and subscripts. 0x2070-0x209F
  {(CHAR16)0x20CF,  1},       // Currency symbols. 0x20A0-0x20CF
  {(CHAR16)0x20FF,  1},       // Combining diacritical marks for symbols. 0x20D0-0x20FF
  {(CHAR16)0x214F,  1},       // Letterlike sympbols. 0x2100-0x214F
  {(CHAR16)0x218F,  1},       // Number forms. 0x2150-0x218F
  {(CHAR16)0x21FF,  1},       // Arrows. 0x2190-0x21FF
  {(CHAR16)0x22FF,  1},       // Mathematical operators. 0x2200-0x22FF
  {(CHAR16)0x23FF,  1},       // Miscellaneous technical. 0x2300-0x23FF
  {(CHAR16)0x243F,  1},       // Control pictures. 0x2400-0x243F
  {(CHAR16)0x245F,  1},       // Optical character recognition. 0x2440-0x245F
  {(CHAR16)0x24FF,  1},       // Enclosed alphanumerics. 0x2460-0x24FF
  {(CHAR16)0x257F,  1},       // Box drawing. 0x2500-0x257F
  {(CHAR16)0x259F,  1},       // Block elements. 0x2580-0x259F
  {(CHAR16)0x25FF,  1},       // Geometric shapes. 0x25A0-0x25FF
  {(CHAR16)0x26FF,  1},       // Miscellaneous symbols. 0x2600-0x26FF
  {(CHAR16)0x27BF,  1},       // Dingbats. 0x2700-0x27BF
  {(CHAR16)0x2FFF,  0},       // Reserved. 0x27C0-0x2FFF
  *
  */

  //
  // CJK phonetics and symbol area
  //
  {(CHAR16)0x33FF,  2},
  /*
   * Merge the blocks and replace them with the above entry as they fall to
   * the same category and they are all wide glyph. This will reduce search
   * time and table size. The merge will omit the reserved code.
   *
   * Remove the above item if below is un-commented.
   *
  {(CHAR16)0x303F,  2},       // CJK symbols and punctuation. 0x3000-0x303F
  {(CHAR16)0x309F,  2},       // Hiragana. 0x3040-0x309F
  {(CHAR16)0x30FF,  2},       // Katakana. 0x30A0-0x30FF
  {(CHAR16)0x312F,  2},       // Bopomofo. 0x3100-0x312F
  {(CHAR16)0x318F,  2},       // Hangul compatibility jamo. 0x3130-0x318F
  {(CHAR16)0x319F,  2},       // Kanbun. 0x3190-0x319F
  {(CHAR16)0x31FF,  0},       // Reserved. As Bopomofo extended in ver3.0. 0x31A0-0x31FF
  {(CHAR16)0x32FF,  2},       // Enclosed CJK letters and months. 0x3200-0x32FF
  {(CHAR16)0x33FF,  2},       // CJK compatibility. 0x3300-0x33FF
  *
  */

  //
  // CJK ideograph area
  //
  {(CHAR16)0x9FFF,  2},
  /*
   * Merge the blocks and replace them with the above entry as they fall to
   * the same category and they are all wide glyph. This will reduce search
   * time and table size. The merge will omit the reserved code.
   *
   * Remove the above item if below is un-commented.
   *
  {(CHAR16)0x4DFF,  0},       // Reserved. 0x3400-0x4DBF as CJK unified ideographs
          // extension A in ver3.0. 0x3400-0x4DFF
  {(CHAR16)0x9FFF,  2},       // CJK unified ideographs. 0x4E00-0x9FFF
  *
  */

  //
  // Reserved
  //
  {(CHAR16)0xABFF,  0},       // Reserved. 0xA000-0xA490 as Yi syllables. 0xA490-0xA4D0
  // as Yi radicals in ver3.0. 0xA000-0xABFF
  //
  // Hangul syllables
  //
  {(CHAR16)0xD7FF,  2},
  /*
   * Merge the blocks and replace them with the above entry as they fall to
   * the same category and they are all wide glyph. This will reduce search
   * time and table size. The merge will omit the reserved code.
   *
   * Remove the above item if below is un-commented.
   *
  {(CHAR16)0xD7A3,  2},       // Hangul syllables. 0xAC00-0xD7A3
  {(CHAR16)0xD7FF,  0},       // Reserved. 0xD7A3-0xD7FF
  *
  */

  //
  // Surrogates area
  //
  {(CHAR16)0xDFFF,  0},       // Surrogates, not used now. 0xD800-0xDFFF

  //
  // Private use area
  //
  {(CHAR16)0xF8FF,  0},       // Private use area. 0xE000-0xF8FF

  //
  // Compatibility area and specials
  //
  {(CHAR16)0xFAFF,  2},       // CJK compatibility ideographs. 0xF900-0xFAFF
  {(CHAR16)0xFB4F,  1},       // Alphabetic presentation forms. 0xFB00-0xFB4F
  {(CHAR16)0xFDFF,  1},       // Arabic presentation forms-A. 0xFB50-0xFDFF
  {(CHAR16)0xFE1F,  0},       // Reserved. As variation selectors in ver3.0. 0xFE00-0xFE1F
  {(CHAR16)0xFE2F,  1},       // Combining half marks. 0xFE20-0xFE2F
  {(CHAR16)0xFE4F,  2},       // CJK compatibility forms. 0xFE30-0xFE4F
  {(CHAR16)0xFE6F,  1},       // Small Form Variants. 0xFE50-0xFE6F
  {(CHAR16)0xFEFF,  1},       // Arabic presentation forms-B. 0xFE70-0xFEFF
  {(CHAR16)0xFFEF,  1},       // Half width and full width forms. 0xFF00-0xFFEF
  {(CHAR16)0xFFFF,  0},       // Speicials. 0xFFF0-0xFFFF
};

STATIC
UINTN
FastGetGlyphWidth (
  IN CHAR16  UnicodeChar
  )
{
  UINTN                         Index;
  UINTN                         Low;
  UINTN                         High;
  UNICODE_WIDTH_ENTRY           *Item;
  UNICODE_WIDTH_ENTRY           *UnicodeWidthTable;

  Item  = NULL;
  Low   = 0;
  UnicodeWidthTable = mFBUnicodeWidthTable;
  High  = (sizeof (mFBUnicodeWidthTable)) / (sizeof (UNICODE_WIDTH_ENTRY)) - 1;
  while (Low <= High) {
    Index = (Low + High) >> 1;
    Item  = &(UnicodeWidthTable[Index]);
    if (Index == 0) {
      if (UnicodeChar <= Item->WChar) {
        break;
      }
      return 0;
    }

    if (UnicodeChar > Item->WChar) {
      Low = Index + 1;
    } else if (UnicodeChar <= UnicodeWidthTable[Index - 1].WChar) {
      High = Index - 1;
    } else {
      //
      // Index - 1 < UnicodeChar <= Index. Found
      //
      break;
    }
  }

  if (Low <= High) {
    return Item->Width;
  }

  return 0;
}

STATIC
EFI_FONT_DISPLAY_INFO *
CreateTtfStringInfo (
  UINT16      FontSize
  )
{
  if (mStringInfo != NULL) {
    mStringInfo->FontInfo.FontSize = (UINT16) FontSize;
    return mStringInfo;
  }

  mStringInfo = AllocateZeroPool (sizeof (EFI_FONT_DISPLAY_INFO) + StrSize (L"ttf") - sizeof (CHAR16));
  ASSERT (mStringInfo != NULL);
  if (mStringInfo == NULL) {
    return NULL;
  }
  mStringInfo->FontInfo.FontSize = (UINT16) FontSize;
  StrCpy (mStringInfo->FontInfo.FontName, L"ttf");

  return mStringInfo;
}

STATIC
FONT_WIDTH_NODE *
GetFontWidthNode (
  IN  CONST UINT16            FontSize,
  IN  CONST FONT_WIDTH_LIST   *FontWidthList
  )
{
  INTN                          Index;

  ASSERT (FontWidthList != NULL);

  Index = FontWidthList->NodeCount;
  while (--Index >= 0) {
    if (FontSize == FontWidthList->FontWidthNode[Index].FontSize) {
      return &FontWidthList->FontWidthNode[Index];
    }
  }
  return NULL;
}

STATIC
FONT_WIDTH_NODE *
SetFontWidthNode (
  IN  CONST UINT16            FontSize,
  IN  FONT_WIDTH_LIST         *FontWidthList
  )
{
  UINT32        NodeCount;

  ASSERT (FontWidthList != NULL);
  NodeCount = FontWidthList->NodeCount;
  FontWidthList->FontWidthNode = ReallocatePool (
                                   NodeCount * sizeof (FONT_WIDTH_NODE),
                                   (NodeCount + 1) * sizeof (FONT_WIDTH_NODE),
                                   FontWidthList->FontWidthNode
                                   );
  SetMem (&FontWidthList->FontWidthNode[NodeCount], sizeof (FONT_WIDTH_NODE), 0xFF);
  FontWidthList->FontWidthNode[NodeCount].FontSize = FontSize;
  FontWidthList->NodeCount = NodeCount + 1;
  return &FontWidthList->FontWidthNode[NodeCount];
}


UINTN
FBGetGlyphWidth (
  UINTN     FontSize,
  CHAR16    Char
  )
{
  EFI_STATUS                 Status;
  UINTN                      FontWidth;
  EFI_FONT_DISPLAY_INFO      *StringInfo;
  EFI_IMAGE_OUTPUT           *Image;
  STATIC FONT_WIDTH_LIST     FontWidthList = {0};
  FONT_WIDTH_NODE            *FontWidthNode;

  ASSERT (FontSize != 0);

  FontWidthNode = NULL;
  if (Char <= 0x00FF) {
    FontWidthNode = GetFontWidthNode ((UINT16) FontSize, &FontWidthList);
    if (FontWidthNode != NULL && FontWidthNode->FontWidth[Char] != 0xFF) {
      return FontWidthNode->FontWidth[Char];
    }

    if (FontWidthNode == NULL) {
      FontWidthNode = SetFontWidthNode ((UINT16) FontSize, &FontWidthList);
      if (FontWidthNode == NULL) {
        return 0;
      }
    }
  } else {
    if (FastGetGlyphWidth (Char) == 2) {
      return FontSize;
    }
  }
  StringInfo = CreateTtfStringInfo ((UINT16) FontSize);
  if (StringInfo == NULL) {
    return 0;
  }
  Image = NULL;

  Status = gHiiFont->GetGlyph (
                       gHiiFont,
                       Char,
                       StringInfo,
                       &Image,
                       NULL
                       );
  if (Status != EFI_SUCCESS) {
    return 0;
  }

  FontWidth = Image->Width;
  if (Char <= 0x00FF) {
    FontWidthNode->FontWidth[Char] = (UINT8)FontWidth;
  }

  FreePool (Image->Image.Bitmap);
  FreePool (Image);
  return FontWidth;
}

BOOL WINAPI
GetTextExtentPoint32 (
  HDC hdc,    /* handle to DC*/
  LPCWSTR lpszStr,  /* character string*/
  int cchString,    /* number of characters*/
  LPSIZE lpSize    /* string dimensions*/
  )
{
  CHAR16                        Char;
  CHAR16                        *Ch;
  UINTN                         Width;
  UINTN                         FontSize;

  if (lpSize) {
    lpSize->cx = 0;
    lpSize->cy = 0;
  }

  if (!hdc || !lpszStr || !cchString || !lpSize)
    return FALSE;

  Ch = (CHAR16 *) lpszStr;
  if (cchString == -1) {
    cchString = INT_MAX;
  }

  FontSize = hdc->font->FontSize;
  if (FontSize == 0) {
    FontSize = 19;
  }

  Width = 0;
  while (--cchString >= 0 && ((Char = *Ch++) != '\0')) {
    Width += FBGetGlyphWidth (FontSize, Char);
  }

  lpSize->cx = (INT32)Width;
  lpSize->cy = (INT32)FontSize;

  return TRUE;
}

int
GetClipBox (
  HDC   Hdc,
  RECT  *Rc
  )
{

  if (Hdc->region == NULL) {
    GetClientRect(Hdc->hwnd, Rc);

  } else if (Hdc->region->rgn->size != 0) {
    *Rc = *(RECT *)(&Hdc->region->rgn->extents);
  } else {
    ASSERT (0);
    SetRectEmpty (Rc);
  }

  return 0;
}


EFI_STATUS
TrueTypeStringToImage (
  IN CHAR16                          *String,
  IN EFI_GRAPHICS_OUTPUT_BLT_PIXEL   *Blt,
  IN UINT16                          FontSize,
  IN UINT16                          Width,
  IN UINT16                          Height,
  IN UINT16                          Delta,
  IN EFI_GRAPHICS_OUTPUT_BLT_PIXEL   FgColor,
  IN EFI_GRAPHICS_OUTPUT_BLT_PIXEL   BkColor,
  OUT UINT16                         *StringWidth
  )
{

  EFI_STATUS                            Status;
  INTN                                  Y;
  EFI_GRAPHICS_OUTPUT_BLT_PIXEL         *BltBuffer;
  EFI_GRAPHICS_OUTPUT_BLT_PIXEL         *BltBuffer2;
  EFI_GRAPHICS_OUTPUT_BLT_PIXEL         *Bitmap;
  EFI_FONT_DISPLAY_INFO                 *StringInfo;
  EFI_IMAGE_OUTPUT                      *Image;
  UINT16                                RetStringWidth;
  UINTN                                 BltBufferWidth;
  UINT16                                ImageWidth;

  ASSERT (String != NULL);
  ASSERT (Blt != NULL);

  StringInfo = CreateTtfStringInfo (FontSize);
  CopyMem (&StringInfo->ForegroundColor, &FgColor, sizeof (EFI_GRAPHICS_OUTPUT_BLT_PIXEL));
  CopyMem (&StringInfo->BackgroundColor, &BkColor, sizeof (EFI_GRAPHICS_OUTPUT_BLT_PIXEL));

  RetStringWidth = 0;
  BltBuffer = Blt;
  while (*String != 0) {
    Image = NULL;
    Status = gHiiFont->GetGlyph (
                         gHiiFont,
                         *String,
                         StringInfo,
                         &Image,
                         NULL
                         );
    if (Status != EFI_SUCCESS) {
      //
      // avoid status is REPLACE_UNKNOWN_GLYPH
      //
      String++;
      continue;
    }

    ImageWidth = Image->Width;
    if (ImageWidth >= Width) {
      FreePool (Image->Image.Bitmap);
      FreePool (Image);
      break;
    }

    RetStringWidth += ImageWidth;

    //
    // glyph to buffer
    //
    Y              = Image->Height;
    Bitmap         = Image->Image.Bitmap;
    BltBuffer2     = BltBuffer;
    BltBufferWidth = ImageWidth * sizeof (EFI_GRAPHICS_OUTPUT_BLT_PIXEL);
    while (--Y >= 0) {
      CopyMem (BltBuffer2, Bitmap, BltBufferWidth);
      BltBuffer2 += Delta;
      Bitmap     += ImageWidth;
    }

    Width     -= ImageWidth;
    BltBuffer += ImageWidth;
    String++;

    FreePool (Image->Image.Bitmap);
    FreePool (Image);
  }

  *StringWidth = RetStringWidth;

  return EFI_SUCCESS;
}


